/*
 * The contents of this file are subject to the terms 
 * of the Common Development and Distribution License 
 * (the License).  You may not use this file except in
 * compliance with the License.
 * 
 * You can obtain a copy of the license at 
 * https://glassfish.dev.java.net/public/CDDLv1.0.html or
 * glassfish/bootstrap/legal/CDDLv1.0.txt.
 * See the License for the specific language governing 
 * permissions and limitations under the License.
 * 
 * When distributing Covered Code, include this CDDL 
 * Header Notice in each file and include the License file 
 * at glassfish/bootstrap/legal/CDDLv1.0.txt.  
 * If applicable, add the following below the CDDL Header, 
 * with the fields enclosed by brackets [] replaced by
 * you own identifying information: 
 * "Portions Copyrighted [year] [name of copyright owner]"
 * 
 * Copyright 2006 Sun Microsystems, Inc. All rights reserved.
 */
package com.sun.grizzly.jruby;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URLConnection;
import java.nio.channels.SocketChannel;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Locale;
import java.util.logging.Level;

import org.jruby.Ruby;
import org.jruby.RubyException;
import org.jruby.RubyIO;
import org.jruby.exceptions.RaiseException;
import org.jruby.javasupport.JavaEmbedUtils;
import org.jruby.runtime.builtin.IRubyObject;

import com.sun.grizzly.http.SelectorThread;
import com.sun.grizzly.http.SocketChannelOutputBuffer;
import com.sun.grizzly.standalone.StaticResourcesAdapter;
import com.sun.grizzly.tcp.ActionCode;
import com.sun.grizzly.tcp.Adapter;
import com.sun.grizzly.tcp.Request;
import com.sun.grizzly.tcp.Response;
import com.sun.grizzly.tcp.http11.InternalOutputBuffer;
import com.sun.grizzly.util.buf.ByteChunk;
import com.sun.grizzly.util.buf.MessageBytes;

/**
 * Adapter implementation that brige JRuby on Rails with Grizzly.
 *
 * @author TAKAI Naoto
 * @author Jean-Francois Arcand
 */
public class RailsAdapter extends StaticResourcesAdapter 
        implements Adapter{

    private static final String RFC_2616_FORMAT = "EEE, d MMM yyyy HH:mm:ss z";
    
    private static final int RAILS_TOKEN = 1;

    private final String publicDirectory;

    private RubyObjectPool pool = null;

    public RailsAdapter(RubyObjectPool pool) {
        super();
        this.pool = pool;
        this.publicDirectory = pool.getRailsRoot() + "/public";

    }

    public void afterService(Request req, Response res) throws Exception {
        try {
            req.action(ActionCode.ACTION_POST_REQUEST, null);
        } catch (Throwable t) {
            t.printStackTrace();
        } finally {
            // Recycle the wrapper request and response
            req.recycle();
            res.recycle();
        }
    }

    public void fireAdapterEvent(String type, Object data) {
    }

    public void service(Request req, Response res) throws Exception { 
        MessageBytes mb = req.requestURI();
        ByteChunk requestURI = mb.getByteChunk();

        try{
            String uri = requestURI.toString();
            File file = new File(publicDirectory,uri);
            if (file.isDirectory()) {
                uri += "index.html";
                file = new File(file,uri);
            }

            if (file.canRead()) {
                super.service(req, res);      
                return;
            } else {
                serviceRails(req, res);
            }

            res.finish();
        } catch (Exception e) {
            if (SelectorThread.logger().isLoggable(Level.SEVERE)) {
                SelectorThread.logger().log(Level.SEVERE, e.getMessage());
            }

            throw e;
        }
    }

    private SocketChannel getChannel(Response res) {
        SocketChannelOutputBuffer buffer = (SocketChannelOutputBuffer) res.getOutputBuffer();
        SocketChannel channel = buffer.getChannel();

        return channel;
    }

    
    private boolean modifiedSince(Request req, File file) {
        try {
            String since = req.getMimeHeaders().getHeader("If-Modified-Since");
            if (since == null) {
                return false;
            }

            Date date = new SimpleDateFormat(RFC_2616_FORMAT, Locale.US).parse(since);
            if (date.getTime() > file.lastModified()) {
                return true;
            } else {
                return false;
            }
        } catch (ParseException e) {
            return false;
        }
    }


    private void serviceRails(Request req, Response res) throws IOException {
        Ruby runtime = null;            
        RailsToken rt = (RailsToken)req.getNote(RAILS_TOKEN);
        if (rt == null){
            rt = new RailsToken();
        }
        rt.req = req;        
        try {
            runtime = pool.bollowRuntime();
            
            if (runtime == null) {
                res.setStatus(503);
                return;
            }
           
            req.doRead(rt.readChunk);
            ((InternalOutputBuffer)res.getOutputBuffer()).commit();
            res.setCommitted(true);

            IRubyObject reqObj = JavaEmbedUtils.javaToRuby(runtime, req);
            IRubyObject loggerObj = JavaEmbedUtils.javaToRuby(runtime, SelectorThread.logger());

            OutputStream os = 
                ((InternalOutputBuffer)res.getOutputBuffer()).getOutputStream();
            
            RubyIO iObj = new RubyIO(runtime, rt.inputStream);
            RubyIO oObj = new RubyIO(runtime, os);

            runtime.defineReadonlyVariable("$req", reqObj);
            runtime.defineReadonlyVariable("$stdin", iObj);
            runtime.defineReadonlyVariable("$stdout", oObj);
            runtime.defineReadonlyVariable("$logger", loggerObj);

            runtime.getLoadService().load("dispatch.rb");
        } catch (RaiseException e) {
            RubyException exception = e.getException();

            System.err.println(e.getMessage());
            exception.printBacktrace(System.err);

            throw e;
        } finally {
            rt.recycle();
            req.setNote(RAILS_TOKEN,rt);
            if (runtime != null) {
                pool.returnRuntime(runtime);
            }
        }
    }
    
    private class RailsInputStream extends InputStream{
        
        public RailsToken rt;
        
        public int read() throws IOException {
            return rt.readChunk.substract();
        }
        
        public int read(byte[] b) throws IOException {
            return read(b,0,b.length);
        } 
        
        public int read(byte[] b, int off, int len) throws IOException {
            return rt.readChunk.substract(b,off,len);
        }
    }
    
    /**
     * Statefull token used to share information with the Rails runtime
     * in a thread safe manner.
     */
    private class RailsToken implements ByteChunk.ByteInputChannel{
        
        public ByteChunk readChunk;

        public Request req;

        public RailsInputStream inputStream;      
             
        public RailsToken(){
            readChunk = new ByteChunk();
            readChunk.setByteInputChannel(this);
            readChunk.setBytes(new byte[8192],0,8192);

            inputStream = new RailsInputStream();
            inputStream.rt = this;
        }
           
        public int realReadBytes(byte[] b, int off, int len) throws IOException {
            req.doRead(readChunk);
            return readChunk.substract(b,off,len);
        }
            
        public void recycle(){
            req = null;
            readChunk.recycle();
        }
        
    }

    @Override
    public String getContentType(String ext) {
        return URLConnection.guessContentTypeFromName("dummy." + ext); 
    }

}
